#ifndef	lint
static char *	Sccsid = "%Z%%M% %I% %G% Spam";
#endif

/* %Z% find standard resistor combination to make desired resistance */

#include	<stdio.h>
#include	<string.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<math.h>

#define		NTRIES		3		/* default tries */

#define		DECADES		8		/* 8 decades total */
#define		SMAX		192		/* max resistors in series */

#define		MAXRES		22000001.	/* maximum resistance */
#define		BIGNUM		999999999.	/* max number */

#define		FALSE		0		/* boolean `false' */
#define		TRUE		1		/* boolean `true' */

#define		ARGCHAR		'-'		/* argument delimiter */
#define		PATHSEP		'/'		/* path separator */

#define		MATCH(x, y)	(!strcmp(x, y))
#define		GETINT(s, x)	(sscanf(s, Int, x) == 1)
#define		GETREAL(s, x)	(sscanf(s, Real, x) == 1)

struct sname {
	char *		name;		/* series name */
	double *	values;		/* series values */
	int		places;		/* precision */
	double		tolerance;	/* tolerance */
	int		expand;		/* expand series by decades */
	};

typedef struct sname	sname_t;

static int	compare(double target, double test);
static int	getevalue(char * string, double * vp);
static void	makeres(void);
static void	doshowvalues(void);
static void	doseries(void);
static void	doparallel(void);
static void	dofeedback(void);
static void	putvalue(double value, int exact);
static void	printeng(double n, int places);
static void	printcode(double n, int places);
static void	getseries(char * name);
static void	setprog(char * name);
static void	dprint(char * fmt, ...);
static void	error(char * fmt, ...);
static void	whyerror(char * fmt, ...);
static void *	mkbuf(int size);
static double	findnext(double start, double target);
static double	parallel(double r1, double r2);
static double	findseries(double r1);
static double	findparallel(double r1);
static double	getmult(int ch);

static double	e3[] = { 1.0, 2.2, 4.7, 0 };

static double	e6[] = { 1.0, 1.5, 2.2, 3.3, 4.7, 6.8, 0 };

static double	e12[] = {
	1.0, 1.2, 1.5, 1.8, 2.2, 2.7, 3.3, 3.9, 4.7, 5.6, 6.8, 8.2, 0 };

static double	e24[] = {
	1.0, 1.1, 1.2, 1.3, 1.5, 1.6, 1.8, 2.0, 2.2, 2.4, 2.7, 3.0,
	3.3, 3.6, 3.9, 4.3, 4.7, 5.1, 5.6, 6.2, 6.8, 7.5, 8.2, 9.1, 0 };

static double	e48[] = {
	1.00, 1.05, 1.10, 1.15, 1.21, 1.27, 1.33, 1.40, 1.47, 1.54, 1.62, 1.69,
	1.78, 1.87, 1.96, 2.05, 2.15, 2.26, 2.37, 2.49, 2.61, 2.74, 2.87, 3.01,
	3.16, 3.32, 3.48, 3.65, 3.83, 4.02, 4.22, 4.42, 4.64, 4.87, 5.11, 5.36,
	5.62, 5.90, 6.19, 6.49, 6.81, 7.15, 7.50, 7.87, 8.25, 8.66, 9.09, 9.53,
	0 };

static double	e96[] = {
	1.00, 1.02, 1.05, 1.07, 1.10, 1.13, 1.15, 1.18, 1.21, 1.24, 1.27, 1.30,
	1.33, 1.37, 1.40, 1.43, 1.47, 1.50, 1.54, 1.58, 1.62, 1.65, 1.69, 1.74,
	1.78, 1.82, 1.87, 1.91, 1.96, 2.00, 2.05, 2.10, 2.15, 2.21, 2.26, 2.32,
	2.37, 2.43, 2.49, 2.55, 2.61, 2.67, 2.74, 2.80, 2.87, 2.94, 3.01, 3.09,
	3.16, 3.24, 3.32, 3.40, 3.48, 3.57, 3.65, 3.74, 3.83, 3.92, 4.02, 4.12,
	4.22, 4.32, 4.42, 4.53, 4.64, 4.75, 4.87, 4.99, 5.11, 5.23, 5.36, 5.49,
	5.62, 5.76, 5.90, 6.04, 6.19, 6.34, 6.49, 6.65, 6.81, 6.98, 7.15, 7.32,
	7.50, 7.68, 7.87, 8.06, 8.25, 8.45, 8.66, 8.87, 9.09, 9.31, 9.53, 9.76,
	0 };

static double	e192[] = {
	1.00, 1.01, 1.02, 1.04, 1.05, 1.06, 1.07, 1.09, 1.10, 1.11, 1.13, 1.14,
	1.15, 1.17, 1.18, 1.20, 1.21, 1.23, 1.24, 1.26, 1.27, 1.29, 1.30, 1.32,
	1.33, 1.35, 1.37, 1.38, 1.40, 1.42, 1.43, 1.45, 1.47, 1.49, 1.50, 1.52,
	1.54, 1.56, 1.58, 1.60, 1.62, 1.64, 1.65, 1.67, 1.69, 1.72, 1.74, 1.76,
	1.78, 1.80, 1.82, 1.84, 1.87, 1.89, 1.91, 1.93, 1.96, 1.98, 2.00, 2.03,
	2.05, 2.08, 2.10, 2.13, 2.15, 2.18, 2.21, 2.23, 2.26, 2.29, 2.32, 2.34,
	2.37, 2.40, 2.43, 2.46, 2.49, 2.52, 2.55, 2.58, 2.61, 2.64, 2.67, 2.71,
	2.74, 2.77, 2.80, 2.84, 2.87, 2.91, 2.94, 2.98, 3.01, 3.05, 3.09, 3.12,
	3.16, 3.20, 3.24, 3.28, 3.32, 3.36, 3.40, 3.44, 3.48, 3.52, 3.57, 3.61,
	3.65, 3.70, 3.74, 3.79, 3.83, 3.88, 3.92, 3.97, 4.02, 4.07, 4.12, 4.17,
	4.22, 4.27, 4.32, 4.37, 4.42, 4.48, 4.53, 4.59, 4.64, 4.70, 4.75, 4.81,
	4.87, 4.93, 4.99, 5.05, 5.11, 5.17, 5.23, 5.30, 5.36, 5.42, 5.49, 5.56,
	5.62, 5.69, 5.76, 5.83, 5.90, 5.97, 6.04, 6.12, 6.19, 6.26, 6.34, 6.42,
	6.49, 6.57, 6.65, 6.73, 6.81, 6.90, 6.98, 7.06, 7.15, 7.23, 7.32, 7.41,
	7.50, 7.59, 7.68, 7.77, 7.87, 7.96, 8.06, 8.16, 8.25, 8.35, 8.45, 8.56,
	8.66, 8.76, 8.87, 8.98, 9.09, 9.20, 9.31, 9.42, 9.53, 9.65, 9.76, 9.88,
	0 };

/* Radio Shack resistor assortment values */

static double	rs[] = {
	1, 2.2,
	10, 15, 22, 33, 39, 47, 51, 68, 82,
	100, 120, 150, 180, 220, 270, 330, 390, 470, 510, 560, 680, 820,
	1000, 1200, 1500, 1800, 2200, 2700, 3000, 3300, 3900, 4700, 5100,
	5600, 6800, 8200,
	10000, 12000, 15000, 18000, 22000, 27000, 33000, 39000, 47000,
	51000, 56000, 68000, 82000,
	100000, 120000, 150000, 180000, 220000, 270000, 330000, 470000,
	1000000, 1500000, 2200000, 3300000, 4700000,
	10000000,
	0 };

/* Radio Shack metal film resistor assortment values */

static double	rsmf[] = {
	10, 47, 100, 150, 220, 470, 1000, 2200, 4700, 10000,
	15000, 100000, 1000000, 0 };

sname_t		snames[] = {
		{ "3",    e3,   2, 50,   TRUE  },
		{ "6",    e6,   2, 20,   TRUE  },
		{ "12",   e12,  2, 10,   TRUE  },
		{ "24",   e24,  2,  5,   TRUE  },
		{ "48",   e48,  3,  2,   TRUE  },
		{ "96",   e96,  3,  1,   TRUE  },
		{ "192",  e192, 3,  0.5, TRUE  },
		{ "rs",   rs,   2,  5,   FALSE },
		{ "rsmf", rsmf, 2,  1,   FALSE },
		{ NULL, NULL, 0, 0 }
		};

static char *	Dcolor[] = {
	"black", "brown", "red", "orange", "yellow",
	"green", "blue", "violet", "grey", "white" };

static char *	Dcolor3[] = {
	"blk", "brn", "red", "org", "yel",
	"grn", "blu", "vio", "gry", "wht" };

static char *	Decolor[] = { "gold" };

static char *	Decolor3[] = { "gld" };

static char *	Usage       =
	"usage: %s [-cfis3] [-np #] [-e <3,6,12,24,48,96,192,rs,rsmf>] [#]";

static char *	Dprefixes   = "munp???";
static char *	Mprefixes   = "KMGT???";
static char *	Real        = "%lf";
static char *	Int         = "%i";
static char *	Dseries     = "12";

static int	debugflag;	/* debug flag */
static int	ntries;		/* # of tries for best match */
static int	places;		/* places of precision */
static int	expand;		/* expand series by decades */
static int	colour;		/* print resistor colour codes */
static int	showint;	/* show integer values */
static char *	prog;		/* program name */
static char **	dcolor;		/* digit colors */
static char **	decolor;	/* digit extended colors */
static double	desired;	/* desired value */
static double	tolerance;	/* tolerance */
static double *	series;		/* resistor series */
static double *	values;		/* resistor values */
static double *	vend;		/* resistor value end */

int
main(
	int		argc,		/* argument count */
	char **		argv,		/* argument vector */
	char **		envp)		/* environment pointer */
{
	int		showvalues;	/* show values */
	int		feedback;	/* compute feedback resistors */
	char *		va;		/* value array */
	char *		nvp;		/* needed value pointer */
	char *		cvp;		/* current value pointer */
	char *		evp;		/* end of value pointers */
	double		precision;	/* precision */

	setprog(*argv);

	va         = mkbuf(argc);
	nvp        = va;
	cvp        = va;
	evp        = va + argc;
	desired    = 0;
	precision  = 0;
	ntries     = NTRIES;
	series     = NULL;
	debugflag  = FALSE;
	feedback   = FALSE;
	showvalues = FALSE;
	colour     = FALSE;
	showint    = FALSE;
	dcolor     = Dcolor;
	decolor    = Decolor;

	while (--argc) {
		if (**++argv == ARGCHAR) {
			while (*++*argv) {
				switch (**argv) {
					case 'n':	/* # of tries */
					case 'p':	/* precision */
					case 'e':	/* EIA series */
						if (nvp == evp) {
							error(Usage, prog);
							}
						*nvp++ = **argv;
						break;

					case 'c':	/* print colour codes */
						colour = TRUE;
						break;

					case 'f':	/* compute feedback */
						feedback = TRUE;
						break;

					case 'i':	/* show integers */
						showint = TRUE;
						break;

					case 's':	/* show values */
						showvalues = TRUE;
						break;

					case '3':	/* use 3-letter codes */
						dcolor  = Dcolor3;
						decolor = Decolor3;
						break;

					case 'D':	/* debug */
						debugflag = TRUE;
						break;

					default:
						error(Usage, prog);
					}
				}
			}
		else {
			if (nvp > cvp) {
				switch (*cvp++) {
					case 'n':	/* # of tries */
						if (!GETINT(*argv, &ntries)) {
							error(Usage, prog);
							}

						break;

					case 'p':	/* precision */
						if (!GETREAL(*argv,
						    &precision)) {
							error(Usage, prog);
							}
						break;

					case 'e':	/* EIA series */
						getseries(*argv);
						break;
					}
				}
			else {
				if (!getevalue(*argv, &desired)) {
					error(Usage, prog);
					}
				}
			}
		}

	if (nvp > cvp) {
		error(Usage, prog);
		}

	free(va);

	if (series == NULL) {
		getseries(Dseries);
		}

	makeres();

	if (showvalues) {
		doshowvalues();
		return	0;
		}

	if (feedback) {
		dofeedback();
		return	0;
		}

	if (desired == 0) {
		error(Usage, prog);
		}

	if (precision != 0) {
		tolerance = precision;
		}

	doparallel();
	doseries();

	if (expand) {
		free(values);
		}

	return	0;
}

/* build values of series */

static void
makeres(void)
{
	int		decade;	/* decade */
	double		value;	/* current value */
	double *	sp;	/* series pointer */

	if (!expand) {
		values = series;

		for (vend = values; *vend != 0; vend++) {
			}

		return;
		}

	values = mkbuf(DECADES * SMAX * sizeof(double));
	vend   = values;
	decade = 10;

	for (;;) {
		for (sp = series; *sp != 0; sp++) {
			value = *sp * decade;

			if (value > MAXRES) {
				return;
				}

			*vend++ = value;
			}

		decade *= 10;
		}
}

/* show values */

static void
doshowvalues(void)
{
	double *	vp;	/* value pointer */

	for (vp = values; vp < vend; vp++) {
		if (showint) {
			printf("%u", (unsigned) (*vp + 0.5));
			}
		else {
			printeng(*vp, places);
			}

		if (colour) {
			printcode(*vp, places);
			}

		printf("\n");
		}
}

/* do series resistance search */

static void
doseries(void)
{
	double		major;	/* major resistance value */
	double		rs;	/* parallel resistance */
	double		rbest;	/* best match */
	double		dbest;	/* best difference */
	double		mbest;	/* best major */
	double		diff;	/* difference */

	/* compute major resistor value */

	major = findnext(0, 0);

	if (major > desired) {
		putvalue(major, FALSE);
		printf(" is already more than ");
		putvalue(desired, FALSE);
		printf(", so a series combination won't help\n");
		return;
		}

	mbest = MAXRES;
	rbest = MAXRES;
	dbest = fabs(desired - (rbest + mbest));

	while (major < desired) {
		rs   = findseries(major);
		diff = fabs(desired - (major + rs));

		dprint("major = %lf, Rs = %lf, diff = %lf (%.2lf%%)",
			major, rs, diff, diff / desired * 100);

		if (diff < dbest) {
			rbest = rs;
			mbest = major;
			dbest = diff;
			}

		major = findnext(major, major);

		if (major == 0) {
			break;
			}
		}

	if (rbest > desired) {
		printf("closest value is ");
		putvalue(rbest, FALSE);
		printf(", which is already more (%.2lf%%)\n",
			fabs(desired - rbest) / desired * 100);
		}
	else {
		putvalue(mbest, FALSE);
		printf(" in series with ");
		putvalue(rbest, FALSE);
		printf(" yields ");
		putvalue(mbest + rbest, TRUE);
		printf(" (%.2lf%%)\n", dbest / desired * 100);
		}
}

/* find best series resistance */

static double
findseries(
	double		r1)	/* series resistance */
{
	double		rbest;	/* best 2nd resistor value */
	double		pdiff;	/* parallel combo difference */
	double		pbest;	/* best parallel combo */
	double *	vp;	/* value pointer */

	pbest = BIGNUM;
	rbest = MAXRES;

	for (vp = values; vp < vend; vp++) {
		pdiff = fabs(desired - (r1 + *vp));

		if (pdiff < pbest) {
			pbest = pdiff;
			rbest = *vp;
			}
		}

	return	rbest;
}

/* do parallel resistance search */

static void
doparallel(void)
{
	double		major;	/* major resistance value */
	double		rp;	/* parallel resistance */
	double		rbest;	/* best match */
	double		dbest;	/* best difference */
	double		mbest;	/* best major */
	double		diff;	/* difference */

	/* compute major resistor value */

	major = findnext(0, desired);

	if (major == 0) {
		printf("larger than largest value, paralleling won't help.\n");
		return;
		}

	/* see if it will do by itself */

	if (compare(desired, major)) {
		putvalue(major, FALSE);
		printf(" is a standard value\n");
		}

	dbest = BIGNUM;
	rp    = MAXRES;

	while (rp > major) {
		rp   = findparallel(major);
		diff = fabs(desired - parallel(major, rp));

		dprint("major = %lf, Rp = %lf, diff = %lf (%.2lf%%)",
			major, rp, diff, diff / desired * 100);

		if (diff < dbest) {
			rbest = rp;
			dbest = diff;
			mbest = major;
			}

		major = findnext(major, major);
		}

	putvalue(mbest, FALSE);
	printf(" paralleled with ");
	putvalue(rbest, FALSE);
	printf(" yields ");
	putvalue(parallel(mbest, rbest), TRUE);
	printf(" (%.2lf%%)\n", dbest / desired * 100);
}

/* find best parallel resistance */

static double
findparallel(
	double		r1)	/* resistance to parallel */
{
	double		rbest;	/* best 2nd resistor value */
	double		pdiff;	/* parallel combo difference */
	double		pbest;	/* best parallel combo */
	double *	vp;	/* value pointer */

	pbest = BIGNUM;

	for (vp = values; vp < vend; vp++) {
		pdiff = fabs(desired - parallel(r1, *vp));

		if (pdiff < pbest) {
			pbest = pdiff;
			rbest = *vp;
			}
		}

	return	rbest;
}

/* compute resistance of two resistors in parallel */

static double
parallel(
	double	r1,	/* 1st resistance */
	double	r2)	/* 2nd resistance */
{
	return	(r1 * r2) / (r1 + r2);
}

/* find resistors for voltage regulator feedback */

static void
dofeedback(void)
{
	double		r1;		/* lower (grounded) resistor */
	double		r2;		/* desired r2 (upper) resistor */
	double		rtest;		/* test resistance */
	double		r1best;		/* best r1 */
	double		r2best;		/* best r2 */
	double		dbest;		/* best difference */
	double		diff;		/* difference */
	double		voltage;	/* voltage */

	dbest = desired;

	/* loop through possible r1 values */

	for (r1 = findnext(0, 240); r1 <= 1500; r1 = findnext(r1, r1)) {

		/* compute optimal r2 for this r1 */

		r2 = r1 * ((desired / 1.21) - 1);

		dprint("r1 = %lf, r2 = %lf", r1, r2);

		for (rtest = findnext(0, 0); rtest != 0;
		     rtest = findnext(rtest, rtest)) {

			voltage = 1.21 * (1 + (rtest / r1));

			diff = fabs(voltage - desired);

			dprint("rtest = %lf, voltage = %lf, diff = %lf",
				rtest, voltage, diff);

			if (diff < dbest) {
				dprint("we have a winner");
				r1best = r1;
				r2best = rtest;
				dbest  = diff;
				}
			}
		}

	printf("r1 = ");
	putvalue(r1best, FALSE);
	printf("; r2 = ");
	putvalue(r2best, FALSE);
	printf(" voltage = ");
	putvalue(1.21 * (1 + (r2best / r1best)), TRUE);
	printf("\n");
}

/* compare two values */

static int
compare(
	double	target,		/* target value */
	double	test)		/* test value */
{
	return	(fabs(target - test) < (target * tolerance / 100));
}

/* find next higher value */

static double
findnext(
	double		start,		/* starting value */
	double		target)		/* target value */
{
	double *	vp;		/* value pointer */

	for (vp = values; vp < vend; vp++) {
		if ((*vp > start) && (*vp >= target)) {
			return	(*vp);
			}
		}

	return	0;
}

/* print resistance value in engineering notation */

static void
putvalue(
	double	value,	/* value to print */
	int	exact)	/* exact flag */
{
	printeng(value, exact ? 4 : places);

	if (colour && !exact) {
		printcode(value, places);
		}
}

/* print resistor colour code */

static void
printcode(
	double	n,		/* number to print */
	int	places)		/* places of accuracy */
{
	int	decade;		/* decade */
	int	digit;		/* current digit */
	double	exponent;	/* exponent */

	printf(" (");

	decade   = 0;
	exponent = pow(10.0, (double) places) - 0.1;

	dprint("n = %lf, exponent = %lf", n, exponent);

	while (n < exponent) {
		n      *= 10;
		decade -= 1;
		}

	while (n > exponent) {
		n      /= 10;
		decade += 1;
		}

	exponent = pow(10.0, (double) (places - 1));

	dprint("n = %lf, decade = %d", n, decade);

	while (places--) {
		digit = n / exponent;

		printf("%s, ", dcolor[digit]);

		n        -= digit * exponent;
		exponent /= 10;
		}

	if (decade < 0) {
		printf("%s", decolor[-decade - 1]);
		}
	else {
		printf("%s", dcolor[decade]);
		}

	printf(") ");
}

/* print number in engineering notation */

static void
printeng(
	double	n,		/* number to print */
	int	places)		/* places of accuracy */
{
	int	decade;		/* decade */
	int	place;		/* current digit place */
	int	digit;		/* current digit */
	double	exponent;	/* exponent */

	decade = 0;

	while (n < 1) {
		n      *= 1000;
		decade -= 3;
		}

	while (n >= 1000) {
		n      /= 1000;
		decade += 3;
		}

	place    = log10(n);
	exponent = pow(10.0, (double) place);

	n += pow(10.0, place - places);

	while (places--) {
		digit = n / exponent;

		printf("%d", digit);

		if ((place == 0) && (places > 0)) {
			printf(".");
			}

		n        -= digit * exponent;
		exponent /= 10;
		place--;
		}

	while (place-- >= 0) {
		printf("0");
		}

	if (decade != 0) {
		if (decade < 0) {
			printf("%c", Dprefixes[-decade / 3 - 1]);
			}
		else {
			printf("%c", Mprefixes[decade / 3 - 1]);
			}
		}
}

/* get value in engineering notation */

static int
getevalue(
	char *		string,	/* string containing value */
	double *	vp)	/* value pointer */
{
	char		uc;	/* units character */
	double		mult;	/* multiplier */

	switch (sscanf(string, "%lf%c", vp, &uc)) {
		case 1:
			return	TRUE;

		case 2:
			mult = getmult(uc);

			if (mult == 0) {
				return	FALSE;
				}

			*vp *= mult;
			return	TRUE;
		}

	return	FALSE;
}

/* look up multiplier value by engineering code */

static double
getmult(
	int	ch)		/* character to look up */
{
	int	exponent;	/* multiplier exponent */
	char *	cp;		/* character pointer */

	for (cp = Mprefixes; *cp; cp++) {
		if (ch == *cp) {
			return	pow(10.0, 3 * (cp - Mprefixes + 1));
			}
		}

	for (cp = Dprefixes; *cp; cp++) {
		if (ch == *cp) {
			return	pow(10.0, -3 * (cp - Dprefixes + 1));
			}
		}

	return	0;
}

/* get series pointer by name */

static void
getseries(
	char *		name)	/* name of series to get */
{
	sname_t *	snp;	/* series name pointer */

	for (snp = snames; snp->name != NULL; snp++) {
		if (MATCH(name, snp->name)) {
			series    = snp->values;
			places    = snp->places;
			tolerance = snp->tolerance / 10;
			expand    = snp->expand;
			return;
			}
		}

	error(Usage, prog);
}

/* set currently running program name */

static void
setprog(
	char *	name)	/* program name to use */
{
	prog = strrchr(name, PATHSEP);

	if (prog) {
		prog++;
		}
	else {
		prog = name;
		}
}

/* print out message (if enabled) */

static void
dprint(
	char *	fmt,	/* message format */
		...)	/* format arguments */
{
	va_list	ap;	/* argument pointer */

	if (debugflag == FALSE) {
		return;
		}

	fflush(stdout);

	va_start(ap, fmt);

	vfprintf(stderr, fmt, ap);

	va_end(ap);

	fprintf(stderr, "\n");
}

/* print out message and exit */

static void
error(
	char *	fmt,	/* message format */
		...)	/* format arguments */
{
	va_list	ap;	/* argument pointer */

	fflush(stdout);

	fprintf(stderr, "%s: ", prog);

	va_start(ap, fmt);

	vfprintf(stderr, fmt, ap);

	va_end(ap);

	fprintf(stderr, "\n");

	exit(EINVAL);
}

/* print out error message and reason and exit */

static void
whyerror(
	char *	fmt,	/* message format */
		...)	/* format arguments */
{
	char *	errstr;	/* error string pointer */
	va_list	ap;	/* argument pointer */

	fflush(stdout);

	fprintf(stderr, "%s: ", prog);

	va_start(ap, fmt);

	vfprintf(stderr, fmt, ap);

	va_end(ap);

	errstr = strerror(errno);

	if (errstr) {
		fprintf(stderr, ": %s\n", errstr);
		}
	else {
		fprintf(stderr, ": errno=%d\n", errno);
		}

	if (errno) {
		exit(errno);
		}
	else {
		exit(EINVAL);
		}
}

/* allocate a buffer and punt on failure */

static void *
mkbuf(
	int	size)		/* size of buffer to allocate */
{
	void *	ptr;		/* address of new buffer */

	ptr = malloc(size);

	if (ptr == NULL) {
		whyerror("Malloc(%d)", size);
		}

	return	ptr;
}
